from pylab import *
from numpy import *
import netCDF4 as nc
import time
import sys
import os
import glob
import pickle
import parcels
from collections import defaultdict

# Missing features so far: The write once variables
# The handling of missing records in *.npy files with "theBadValues" must be improved
#
# assumes dictionaries are ordered, so python 3.7 or better.
#
#
# The goal of this code is to convert from *.npy temporary files to
# netCDF without using too much memory. The memory ussage of this code
# will scale as (the number of time records in longest drifter record)
# multiplied by (the number of drifters in a single MPI rank). Thus it
# will use less memory than the existing code, in which memory ussage
# scales as the (number of drifters)*(number of output times). In
# paticular, the memory ussage will decline when MPI is used, roughly
# as 1/(number of MPI jobs). Memory useage will decline if drifters
# have a finite duration (i.e. they die after some time), because that
# reduces the number of time records in the longest drifter record.
#
# The downside of this scheme is all data must be read twice, in order
# to figure out the temporal extent of each drifter record.
#

printDebug=True #if true print lots of debugging info

assert sys.version_info.major>=3 and sys.version_info.minor>=7,'Needs python 3.7 or greater to run' 

#set or get names of input and output files
if True: #get from command line
    if len(sys.argv)!=3:
        print(' ')
        print(' This script converts from the out-* temporary directories that parcels')
        print(' creates to a netCDF output. ')
        print(' ')
        print(' It must be called with two arguements: python covertToRaggedNetcdf.py outDirectory fileOut')
        print(' It must be called from the directory in which parcels was run')
        print(' ')
        sys.exit(1)
    dirIn=sys.argv[1]
    fileOut=sys.argv[2]
else:
    #just define here
    if True: #big
        dirIn='dataPaths/out-IEVTCOSZ'
        fileOut='dataPaths/test_squareNetCDF.nc'
    elif False:
        #small run
        dirIn='dataPaths/out-DYFGVYPB'
        fileOut='dataPaths/test_squareNetCDF_SMALL.nc'
    elif True:
        #small run
        dirIn='dataPaths/out-XVINWKJA'
        fileOut='dataPaths/test_squareNetCDF_3months_shortDrift.nc'        

#lets try not to destroy data
if os.path.exists(fileOut):
    print(' ')
    print('File %s exists, and will not be overwritten. Please delete if you wish.'%(fileOut,) )
    sys.exit(1)

    
#search for all data directories within the data directories, and record all times in each. 
#get all temporary sub directories, sorted
dirsInData=sorted(glob.glob(os.path.join("%s" % dirIn, "*")),key=lambda x: int(os.path.basename(x)))

#now for each temporary data directory, load the pset_info.npy data
psetInfoDict={}
for tempDir in dirsInData:
    nDir=int(os.path.basename(tempDir))
    psetInfoDict[nDir]=load(os.path.join(tempDir,'pset_info.npy'),allow_pickle=True).item()

#what variables to convert
#NOTE, THIS DOES NOT USE THE INFORMATION IN var_names_once and other *once feilds.... HACK
#ALSO ASSUMES ALL SUBDIRECTORIES HAVE THE SAME FIELDS
onePsetInfoList=psetInfoDict[list(psetInfoDict.keys())[0]]
varList=onePsetInfoList['var_names']
varDtypesList=onePsetInfoList['var_dtypes']

#now we need to scan all the data files for all the ranks to figure
#out how many drifters there are ("numberOfDrifters"), and at what
#times they started and stopped (all times are kept as integer record
#numbers). Also record the maximum number of time records for any
#difter, so we can make the "obs" dimension of the netCDF file the right size. 

#This is wasteful of IO, since all data is read twice. However, it
#conserves memory, since we don't ever keep the full data set in
#memory. And that is the point of this code.

#dictionary with MPI rank as key to a set of dictionaries with drifter
#id as a key that contain the starting and ending time record of each
#drifter. Not that the number of times the drifter output is not
#always drifterTmax-drifterTmin+1, since sometimes the drifter output
#skips an *.npy file for some reason.
drifterTminAll={}
drifterTmaxAll={}
numberOfDrifters=0 #will end up being the total number of drifters in file
maxNumObs=-1 #will end up being the maximum number of observations for any drifter
if printDebug:
    print('   Starting to find time ranges')
for nRank in range(len(psetInfoDict)):
    #in each directory, there is one file for each output time. read
    #each of these files in, and record for each drifter the first and
    #last output file it has data in. Don't save the rest of the data
    tic=time.time()
    drifterTmin={} #first *npy file with the drifter in it. Key is drifter id
    drifterTmax={} #last *npy file with the drifter in it. Key is drifter id
    obsPerDrifter=defaultdict(lambda: 0) #number of observations per drifter. Key is drifter id
    nt=-1
    for f in psetInfoDict[nRank]['file_list']:
        nt+=1 #this assumes the file_list is ordered in time!
        aTime=load(f,allow_pickle=True).item()
        #update driftTmax if id exists at this time. Only update
        #drifterTmin if drifter not already in drifterTmin
        [drifterTmax.update({id:nt}) for id in aTime['id']]
        [obsPerDrifter.update({id:obsPerDrifter[id]+1}) for id in aTime['id']]
        [drifterTmin.update({id:nt}) for id in aTime['id'] if (not (id in drifterTmin))]
    drifterTminAll[nRank]=drifterTmin
    drifterTmaxAll[nRank]=drifterTmax
    numberOfDrifters+=len(drifterTmin)
    maxNumObs=max(maxNumObs,max(obsPerDrifter.values()))
    if printDebug:
        print('         Done with finding time ranges in',time.time()-tic,'for rank',nRank)
        print('         There are',len(drifterTmax),'drifters and maxObs for this rank is',max(obsPerDrifter.values()))
if printDebug:
    print('      There are',numberOfDrifters,'drifters in all ranks')
    print('      The drifter with the most observations has ',maxNumObs,'observations')
    #assert False,'as'

#THIS CODE DOES NOT WORK, BECAUSE SOMETIMES A DRIFTER WILL SKIP A *.npy FILE    
# #to discover the maximum number of times to write out, scan the
# #drifterTminAll and drifterTmaxAll dictionaries to find the maximum
# #number of times any drifter exists
# maxNumObs=-1
# for rankKey in drifterTmaxAll.keys():
#     for drifterKey in drifterTmaxAll[rankKey].keys():
#         maxNumObs=max(drifterTmaxAll[rankKey][drifterKey]-drifterTminAll[rankKey][drifterKey]+1,maxNumObs)
# if printDebug:
#     print('      The maximum number of observations for a drifter are',maxNumObs)
    
#there are some differences in names between the netCDF output and the
#names in the .npy file. In particular, "id" is called
#"trajectory" and "depth" is called "z".
#Make a dictionary that maps from one to the other
varNamesNetCDF=dict(zip(varList,varList))
varNamesNetCDF['id']='trajectory'
varNamesNetCDF['depth']='z'
    
#Create the output netCDF file.
ncfile = nc.Dataset(fileOut,'w',format='NETCDF4')
ncfile.feature_type = "trajectory"
ncfile.Conventions = "CF-1.6/CF-1.7"
ncfile.ncei_template_version = "NCEI_NetCDF_Trajectory_Template_v2.0"
ncfile.parcels_version = parcels.__version__
ncfile.parcels_mesh = onePsetInfoList['parcels_mesh']

#make dims that are fixed
traj_dim=ncfile.createDimension('traj', numberOfDrifters)
#obs_dim=ncfile.createDimension('obs', len(psetInfoDict[0]['file_list']))
obs_dim=ncfile.createDimension('obs', maxNumObs)

#create variables
varFillValue={} # a dictionary of the fill value for each variable. Needed later
nVar=-1
for var in varList:
    nVar+=1
    #define variable
    dtype=psetInfoDict[0]['var_dtypes'][nVar]
    fillVal=NaN if dtype in {float32,float64} else iinfo(dtype).max
    varFillValue[var]=fillVal
    vlvar = ncfile.createVariable(varNamesNetCDF[var], dtype,('traj','obs'),fill_value=fillVal)

    if var=='time':
        vlvar.long_name = ""
        vlvar.standard_name = "time"
        if onePsetInfoList['time_origin'] is None:
            vlvar.units="seconds"
        else:
            vlvar.units="seconds since "+str(onePsetInfoList['time_origin'])
            vlvar.calendar='standard'
    elif var=='lon':
        vlvar.long_name = ""
        vlvar.standard_name = "longitude"
        vlvar.units="degrees_east"
        vlvar.axis="X"
    elif var=='lat':
        vlvar.long_name = ""
        vlvar.standard_name = "latitude"
        vlvar.units="degrees_north"
        vlvar.axis="Y"
    elif var=='depth' or var=='z':
        vlvar.long_name = ""
        vlvar.standard_name = "depth"
        vlvar.units="m"
        vlvar.positive="down"
    else:
        vlvar.long_name = ""
        vlvar.standard_name = var
        vlvar.units="unknown"

#ncfile.close()
#assert False,'as'
        
#now loop over each directory containing data output.  if parcels was
#run with MPI, there should be one directory for each MPI process
nRecOut=0 #this keeps track of how many netCDF records are written out.
weirdTooLong=0 #keep track of an odd anomaly
for nRank in range(len(psetInfoDict)):
    ticOuter=time.time()    
    if printDebug:
        print('starting on nRank',nRank)
    
    drifterTmin=drifterTminAll[nRank]
    drifterTmax=drifterTmaxAll[nRank]
        
    #for efficiecy, get this just once
    var_names=psetInfoDict[nRank]['var_names']
    
    #make a dictionary to store the drifter data from the *.npy
    #files. The trick is to keep this as small as possible at all
    #times. The key to this dictionary will be a time index, and the
    #contents will be the content of a *.npy file from file_list
    dataFromNpy={}

    #make a dictionary to store the data to be saved. When it gets big
    #enough, write it out to netCDF
    driftersToStore={}
    
    #now loop over the times in file_list, starting at the
    #least. Process at drifters that start at thisStartRec until there
    #are no more drifters left that started at thisStartRec. By the
    #time we have done this for all time records, we should have
    #processed all drifters.
    for thisStartRec in range(len(psetInfoDict[nRank]['file_list'])):
        ticInner=time.time()
        #if printDebug:
        #    print('   Processing all drifters that start at time',thisStartRec)

        nDone=-1
        listOfDrifters=list(drifterTmax.keys()) #to avoid issues in removing keys from an iterator
        for driftID in listOfDrifters: #get a drifter
            if drifterTmin[driftID]==thisStartRec: #starts at thisStartRec, so process

                nDone+=1
                
                #use pop, so that we remove the drifter from list of
                #drifters under consideration since we will process it now 
                thisDriftStart=drifterTmin.pop(driftID)
                thisDriftEnd=drifterTmax.pop(driftID)

                #now make arrays to store drifter data in
                thisDrift={}
                nVar=-1
                for var in psetInfoDict[nRank]['var_names']:
                    nVar+=1
                    thisDrift[var]=zeros((thisDriftEnd-thisDriftStart+1,),
                                         dtype=psetInfoDict[nRank]['var_dtypes'][nVar])

                #now loop over time and get data
                for nRec in range(thisDriftStart,thisDriftEnd+1):

                    #find where in npy vectors this difter ID is. If
                    #the npy file is not in dataFromNpy then load it.
                    try: #this will fail if nRec is not in dataFromNpy
                        try:
                            #this try except is because drifters
                            #sometimes skip a *.npy file and are in a
                            #later one. Not sure why this is, but it is.
                            nWhere=dataFromNpy[nRec]['idWhereFrom'][dataFromNpy[nRec]['idSorted'].searchsorted(driftID)]
                            dataIsPresent=True
                        except IndexError:
                            dataIsPresent=False
                            nWhere = len(dataFromNpy[nRec]['idWhereFrom'])+5 #this is bad index
                            
                    except KeyError:
                        #oops, have not loaded this record from *.npy file.
                        dataFromNpy[nRec]=load(psetInfoDict[nRank]['file_list'][nRec],allow_pickle=True).item()

                        #now, to speed up the search for the drifter
                        #id, it is useful to add a sorted version of
                        #the id vector, along with the indices it came
                        #from, so I can use a much faster searchsorted
                        #routine to find nWhere
                        dataFromNpy[nRec]['idWhereFrom']=argsort(dataFromNpy[nRec]['id'])
                        dataFromNpy[nRec]['idSorted']=dataFromNpy[nRec]['id'][dataFromNpy[nRec]['idWhereFrom']]
                        dataIsPresent=True
                        nWhere=dataFromNpy[nRec]['idWhereFrom'][dataFromNpy[nRec]['idSorted'].searchsorted(driftID)]
                        

                    #now save the data to thisDrift. Made more verbose for speed.
                    thisRec=dataFromNpy[nRec]
                    for var in var_names:
                        theBadValue=varFillValue[var] #what to fill missing values with
                        if dataIsPresent:
                            thisDrift[var][nRec-thisDriftStart]=thisRec[var][nWhere]
                        else:
                            #ok, this is a pain in the ass. Sometimes,
                            #an output time for a particle is written
                            #at a later *.npy file. Why, I don't
                            #know. But I must fill the record with bad
                            #value. 
                            thisDrift[var][nRec-thisDriftStart]=theBadValue #this should not happen

                #now store the starting and ending records of this drifter
                thisDrift['nStart']=thisDriftStart
                thisDrift['nEnd']=thisDriftEnd
                
                #now store the data in driftersToStore
                driftersToStore[driftID]=thisDrift

        #now we are done with all drifters that started at this time, so I can remove
        #this time record of data from dataFromNpy
        del dataFromNpy[thisStartRec]
        #if printDebug:
        #    print('      done with thisStartRec',thisStartRec,'in',time.time()-ticInner)

        #now would be a good time to save data in driftersToStore to
        #netCDF and then clear it. Note in this code, the
        #drifter ID's will not be written out in order.
        ticNetcdf=time.time()
        for var in varList:
            #make an array to store data in, filled with the FillValue
            outArray=zeros([len(driftersToStore),ncfile[varNamesNetCDF[var]].shape[1]],
                           dtype=ncfile[varNamesNetCDF[var]].dtype)+ncfile[varNamesNetCDF[var]]._FillValue

            #note that the following code strongly assumes that
            #dictionaries return things in the same order each
            #time... so python 3.7 or better.

            #pack array full of data
            ticJnk=time.time()
            theBadValue=varFillValue[var]

            #here we have an issue. We need to test bad values in the
            #data. these bad values occur when a drifter skips a *.npy
            #file, so there is a gap in the data. Unfortunately, these
            #gaps are represented as NaN's for floating point values,
            #and the maximum integer value for integers. There is no
            #consistent way to check for nan and a number
            #(e.g. nan==nan is False), so we need to have seperate
            #tests for these to cases, depending on what theBadValue
            #is.
            if isnan(theBadValue):
                nOut=-1
                for thing in driftersToStore.items():
                    nOut+=1
                    thing1=thing[1]
                    #only store good values of the data
                    indx=isfinite(thing1[var])
                    #The max in the next line is dodgy...
                    outArray[nOut,:min(maxNumObs,sum(indx))]=thing1[var][indx][:min(maxNumObs,sum(indx))]
                    if printDebug:
                        if sum(indx)>maxNumObs:
                            print('for var',var,'in rank',nRank,'drifter ID',thing[0],'too many data points?')
                            weirdTooLong+=1
                            #time.sleep(3)
            else:
                nOut=-1
                for thing in driftersToStore.items():
                    nOut+=1
                    thing1=thing[1]
                    #only store good values of the data
                    indx=thing1[var]!=theBadValue
                    #The max in the next line is dodgy...
                    outArray[nOut,:min(maxNumObs,sum(indx))]=thing1[var][indx][:min(maxNumObs,sum(indx))]
                    if printDebug:
                        if sum(indx)>maxNumObs:
                            print('for var',var,'in rank',nRank,'drifter ID',thing[0],'too many data points?')
                            weirdTooLong+=1
                            #time.sleep(3)
                    
            #write all the data to a file
            ticJnk=time.time()
            ncfile[varNamesNetCDF[var]][nRecOut:nRecOut+len(driftersToStore),:]=outArray

        #only update nRecOut when all variables have been written
        nRecOut=nRecOut+len(driftersToStore)

        #now that we have written out the data, reset driftersToStore
        #to an empty dictionary to save memory
        driftersToStore={}
            
        #if printDebug:
        #    print('      done with writing netCDF for this rank in',time.time()-ticNetcdf)

    if printDebug:
        print('      done with all records for rank',nRank,' in',time.time()-ticOuter)
    #return None

if weirdTooLong>0:
    print(weirdTooLong//len(varList),'Drifter records appeared to have too many data points?')
ncfile.close()
print('done')



            
                    
